import java.io.FileReader;
import java.io.BufferedReader;
import java.io.IOException;
import java.io.PrintStream;
import javax.swing.*;
import java.lang.Math;
import java.util.Random;

/** 
 * An interface to the robot information.  Given some script file and
 * map file (both will be provided) you will be able to get status
 * information at each step the robot takes.
 * @author Matt Hoffman
 * @version 2.0
 ***/
public class RobotModel
{
    private static Random r = null;
    private static PrintStream err   = System.err;
    private static PrintStream debug = System.err;

    private JFrame mFrame;
    private BufferedReader mScript;
    private GridMap mMap;
    private boolean mDone = false;

    // only look every 20 cm. this is squared because I don't want to
    // have to do a sqrt when I'm calculating whether something is
    // over the threshold
    private final int DIST_THRESHOLD = 200*200;

    /** The initial x-position for the robot. NOTE: this will be
     * removed for the second half of the project. */
    public final int initialX; 

    /** The initial x-position for the robot. NOTE: this will be
     * removed for the second half of the project. */
    public final int initialY;

    /** The initial rotation for the robot. NOTE: this will be removed
     * for the second half of the project. */
    public final float initialR;
    
    /**
     * Construct a robot model object given the path to some script
     * file, and some map file. If the map-file is null, this will
     * give you an open area with no walls to test your code. All file
     * locations should be relative to the current working directory,
     * ie the directory from which the program is launched.
     * @param script The location of script file.
     * @param map The location of the map file.
     ***/
    public RobotModel (String script, String map)
    {
        // create our map
        mMap = new GridMap(map);

        // try to open our file and then discard everything up to the
        // first scan; any errors that occur should be fatal.
        try {
            mScript = new BufferedReader(new FileReader(script));
            initialize();
        } 
        catch (Exception e) {
            err.println("" + e);
            System.exit(-1);
        }

        // XXX: set the initial position variables; this should be
        // removed for the second half of the project
        initialX = 0;
        initialY = 0;
        initialR = 0.0f;

        mFrame = new JFrame("Display");
        mFrame.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
        mFrame.getContentPane().add(mMap);
        mFrame.pack();
        mFrame.setVisible(true);

        debug.println("Initially at (" + mMap.robotX + "," +
                mMap.robotY + ") and " + mMap.robotR + " radians");
        debug.println("--------------------------------");
    }

    private void initialize() throws IOException
    {
        // look ahead until we find the first scan
        while (true) {
            String line = mScript.readLine();

            // if we can't find a first scan the file is invalid
            // somehow.
            if (line == null) 
                throw new IOException("Invalid file-format.");

            if (line.startsWith("#Scan"))
                break;
        }
    }

    /** Generate a gaussian with mean 0.0 and variance 1.0. */
    public static double randGaussian() 
        { return randGaussian(0.0, 1.0); }

    /**
     * Generate a gaussian with mean m and variance s^2.
     * @param m The mean of the distribution.
     * @param s The standard deviation.
     ***/
    public static double randGaussian(double m, double s)
    {
        if (r == null)
            r = new Random(System.currentTimeMillis());
        return (s * r.nextGaussian() + m);
    }
    
    /**
     * Return whether or not the position is occupied. This will tell
     * you whether the robot can move there or not, and whether or not
     * you should put particles there.
     * @param x The x position.
     * @param y The y position.
     ***/
    public boolean occupied(int x, int y) { 
        return mMap.occupied(x,y);
    }

    /** 
     * Add a particle to the map. This will be deleted by the next
     * call to getNext(), as this is only for drawing the particles.
     ***/
    public void addParticle(int x, int y) 
    {
        mMap.addParticle(x,y);
    }

    /** Schedule the map to be redrawn. */
    public void redraw() { 
        mMap.repaint(0, 0, 0, mMap.getWidth(), mMap.getHeight());
    }

    /** 
     * Get the next status-update for the robot, ie its movement and
     * sensor information. This method will return null when the robot
     * is done moving.
     ***/
    public Status getNext()
    {
        if (mDone) return null;
        
        // make a new status object, which we'll fill up
        Status status = new Status();
        int[]   xy = {mMap.robotX, mMap.robotY};
        float[] ro = {mMap.robotR};

        float xdiff = 0.0f;
        float ydiff = 0.0f;
        float dist  = 0.0f;

        // grab the next position, which must be some distance away as
        // specified by the distance threshold (this is actually the
        // squared distance so we don't have to use sqrt).
        try {
            while (dist < DIST_THRESHOLD)
            {
                getNextSignal(xy, ro);
                xdiff = xy[0] - mMap.robotX;
                ydiff = xy[1] - mMap.robotY;
                dist  = xdiff*xdiff + ydiff*ydiff;
            } 
        }
        catch (Exception e) {
            err.println("" + e);
            System.exit(-1);
        }

        // update the status that we can pass back
        status.distance = (float) Math.sqrt(dist);
        status.initialRotation = 
            (float) Math.atan2(ydiff, xdiff) - mMap.robotR;
        status.finalRotation = ro[0] - status.initialRotation 
                                - mMap.robotR;
        status.cleanupRotation();

        mMap.clearParticles();
        mMap.robotR = ro[0];
        mMap.robotX = xy[0];
        mMap.robotY = xy[1];

        debug.println("Located at ("+mMap.robotX+","+mMap.robotY+") "
                      +mMap.robotR);

        return status;
    }

    // we pass in arrays because java doesn't have any way to pass in
    // an integer-ref, etc.
    private void getNextSignal (int[] xy, float[] ro)
        throws IOException
    {
        while (true)
        {
            String line = mScript.readLine();
            if (line == null) { mDone = true; break; }
            if (line.startsWith("#Scan")) break;

            // split the input string
            String[] parts = line.split(":\\s+");

            if (parts[0].equals("robot")) 
            {
                String[] pose = parts[1].split("\\s+");
                xy[0] = Integer.parseInt(pose[0]);
                xy[1] = Integer.parseInt(pose[1]);
                ro[0] = (float)
                    Math.toRadians(Double.parseDouble(pose[2]));
            }
        }
    }
};
